/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
   This software is licensed as OpenSource, under the Apache License, Version 2.0.
   This license is available at: http://opensource.org/licenses/Apache-2.0. */
/***********************************************************************/

#ifndef TYPECOMP_PACKAGE_SPECS_H
#define TYPECOMP_PACKAGE_SPECS_H

#define TYPECOMP "../include/typecomp.h"
#define DYNARR "../include/dynarr.h"
#define PSTOKEN "../include/pstoken.h"
#define DICTOPS "../include/dictops.h"
#define EXCS0 "../include/excs0.h"
#define EXSUBCS0 "../include/exsubcs0.h"
#define ISOCS0 "../include/isocs0.h"
#define STDENC0 "../include/stdenc0.h"
#define STDSTR0 "../include/stdstr0.h"
#define STDSTR1 "../include/stdstr1.h"
#define TXOPS "../include/txops.h"
#if TC_EURO_SUPPORT
#define CFFREAD "../include/cffread.h"
#endif /* TC_EURO_SUPPORT */

#endif /* TYPECOMP_PACKAGE_SPECS_H */
