/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
   This software is licensed as OpenSource, under the Apache License, Version 2.0.
   This license is available at: http://opensource.org/licenses/Apache-2.0. */

DCL_KEY("-|", kRD)
DCL_KEY("BaseFontBlend", kBaseFontBlend)
DCL_KEY("BaseFontName", kBaseFontName)
DCL_KEY("Blend", kBlend)
DCL_KEY("BlendDesignMap", kBlendDesignMap)
DCL_KEY("BlendDesignPositions", kBlendDesignPositions)
DCL_KEY("BlueFuzz", kBlueFuzz)
DCL_KEY("BlueScale", kBlueScale)
DCL_KEY("BlueShift", kBlueShift)
DCL_KEY("BlueValues", kBlueValues)
DCL_KEY("CIDCount", kCIDCount)
DCL_KEY("CIDFontName", kCIDFontName)
DCL_KEY("CIDFontRevision", kCIDFontRevision)
DCL_KEY("CIDFontType", kCIDFontType)
DCL_KEY("CIDFontVersion", kCIDFontVersion)
DCL_KEY("CIDInit", kCIDInit)
DCL_KEY("CIDMapOffset", kCIDMapOffset)
DCL_KEY("Chameleon", kChameleon)
DCL_KEY("CharStrings", kCharStrings)
DCL_KEY("Copyright", kCopyright)
DCL_KEY("Encoding", kEncoding)
DCL_KEY("Erode", kErode)
DCL_KEY("ExpansionFactor", kExpansionFactor)
DCL_KEY("FDArray", kFDArray)
DCL_KEY("FDBytes", kFDBytes)
DCL_KEY("FSType", kFSType)
DCL_KEY("FamilyBlues", kFamilyBlues)
DCL_KEY("FamilyName", kFamilyName)
DCL_KEY("FamilyOtherBlues", kFamilyOtherBlues)
DCL_KEY("FontBBox", kFontBBox)
DCL_KEY("FontMatrix", kFontMatrix)
DCL_KEY("FontName", kFontName)
DCL_KEY("FontType", kFontType)
DCL_KEY("ForceBold", kForceBold)
DCL_KEY("ForceBoldThreshold", kForceBoldThreshold)
DCL_KEY("FullName", kFullName)
DCL_KEY("GDBytes", kGDBytes)
DCL_KEY("GlyphDirectory", kGlyphDirectory)
DCL_KEY("ItalicAngle", kItalicAngle)
DCL_KEY("LanguageGroup", kLanguageGroup)
DCL_KEY("Notice", kNotice)
DCL_KEY("Ordering", kOrdering)
DCL_KEY("OrigFontType", kOrigFontType)
DCL_KEY("OtherBlues", kOtherBlues)
DCL_KEY("PaintType", kPaintType)
DCL_KEY("PostScript", kPostScript)
DCL_KEY("Private", kPrivate)
DCL_KEY("Registry", kRegistry)
DCL_KEY("RndStemUp", kRndStemUp)
DCL_KEY("RunInt", kRunInt)
DCL_KEY("SDBytes", kSDBytes)
DCL_KEY("StdHW", kStdHW)
DCL_KEY("StdVW", kStdVW)
DCL_KEY("StemSnapH", kStemSnapH)
DCL_KEY("StemSnapV", kStemSnapV)
DCL_KEY("StrokeWidth", kStrokeWidth)
DCL_KEY("SubrCount", kSubrCount)
DCL_KEY("SubrMapOffset", kSubrMapOffset)
DCL_KEY("Subrs", kSubrs)
DCL_KEY("Supplement", kSupplement)
DCL_KEY("UIDBase", kUIDBase)
DCL_KEY("UnderlinePosition", kUnderlinePosition)
DCL_KEY("UnderlineThickness", kUnderlineThickness)
DCL_KEY("UniqueID", kUniqueID)
DCL_KEY("WasEmbedded", kWasEmbedded)
DCL_KEY("Weight", kWeight)
DCL_KEY("WeightVector", kWeightVector)
DCL_KEY("XUID", kXUID)
DCL_KEY("hires", khires)
DCL_KEY("initialRandomSeed", kinitialRandomSeed)
DCL_KEY("isFixedPitch", kisFixedPitch)
DCL_KEY("lenIV", klenIV)
DCL_KEY("version", kversion)
