/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
This software is licensed as OpenSource, under the Apache License, Version 2.0.
This license is available at: http://opensource.org/licenses/Apache-2.0. */
/***********************************************************************/

#ifndef MAKEOTF_PACKAGE_SPECS_H
#define MAKEOTF_PACKAGE_SPECS_H

#define TYPECOMP "typecomp.h"
#define HOTCONV "hotconv.h"
#define DYNARR "dynarr.h"

#endif /* MAKEOTF_PACKAGE_SPECS_H */
