/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
   This software is licensed as OpenSource, under the Apache License, Version 2.0.
   This license is available at: http://opensource.org/licenses/Apache-2.0. */

/* Mac OS 9.0 Cyrillic aggregate Unicode initializer.

   Element values are UVs. UV_UNDEF is 0xFFFF. Index by code, get UV.
   Source: ftp://ftp.unicode.org/Public/MAPPINGS/VENDORS/APPLE/UKRAINE.TXT
   as of 9/9/99, with EURO SIGN replacing CURRENCY SIGN. */

UV_UNDEF,   /* 00 */
UV_UNDEF,   /* 01 */
UV_UNDEF,   /* 02 */
UV_UNDEF,   /* 03 */
UV_UNDEF,   /* 04 */
UV_UNDEF,   /* 05 */
UV_UNDEF,   /* 06 */
UV_UNDEF,   /* 07 */
UV_UNDEF,   /* 08 */
UV_UNDEF,   /* 09 */
UV_UNDEF,   /* 0A */
UV_UNDEF,   /* 0B */
UV_UNDEF,   /* 0C */
UV_UNDEF,   /* 0D */
UV_UNDEF,   /* 0E */
UV_UNDEF,   /* 0F */
UV_UNDEF,   /* 10 */
UV_UNDEF,   /* 11 */
UV_UNDEF,   /* 12 */
UV_UNDEF,   /* 13 */
UV_UNDEF,   /* 14 */
UV_UNDEF,   /* 15 */
UV_UNDEF,   /* 16 */
UV_UNDEF,   /* 17 */
UV_UNDEF,   /* 18 */
UV_UNDEF,   /* 19 */
UV_UNDEF,   /* 1A */
UV_UNDEF,   /* 1B */
UV_UNDEF,   /* 1C */
UV_UNDEF,   /* 1D */
UV_UNDEF,   /* 1E */
UV_UNDEF,   /* 1F */
0x0020,     /* 20  SPACE */
0x0021,     /* 21  EXCLAMATION MARK */
0x0022,     /* 22  QUOTATION MARK */
0x0023,     /* 23  NUMBER SIGN */
0x0024,     /* 24  DOLLAR SIGN */
0x0025,     /* 25  PERCENT SIGN */
0x0026,     /* 26  AMPERSAND */
0x0027,     /* 27  APOSTROPHE */
0x0028,     /* 28  LEFT PARENTHESIS */
0x0029,     /* 29  RIGHT PARENTHESIS */
0x002A,     /* 2A  ASTERISK */
0x002B,     /* 2B  PLUS SIGN */
0x002C,     /* 2C  COMMA */
0x002D,     /* 2D  HYPHEN-MINUS */
0x002E,     /* 2E  FULL STOP */
0x002F,     /* 2F  SOLIDUS */
0x0030,     /* 30  DIGIT ZERO */
0x0031,     /* 31  DIGIT ONE */
0x0032,     /* 32  DIGIT TWO */
0x0033,     /* 33  DIGIT THREE */
0x0034,     /* 34  DIGIT FOUR */
0x0035,     /* 35  DIGIT FIVE */
0x0036,     /* 36  DIGIT SIX */
0x0037,     /* 37  DIGIT SEVEN */
0x0038,     /* 38  DIGIT EIGHT */
0x0039,     /* 39  DIGIT NINE */
0x003A,     /* 3A  COLON */
0x003B,     /* 3B  SEMICOLON */
0x003C,     /* 3C  LESS-THAN SIGN */
0x003D,     /* 3D  EQUALS SIGN */
0x003E,     /* 3E  GREATER-THAN SIGN */
0x003F,     /* 3F  QUESTION MARK */
0x0040,     /* 40  COMMERCIAL AT */
0x0041,     /* 41  LATIN CAPITAL LETTER A */
0x0042,     /* 42  LATIN CAPITAL LETTER B */
0x0043,     /* 43  LATIN CAPITAL LETTER C */
0x0044,     /* 44  LATIN CAPITAL LETTER D */
0x0045,     /* 45  LATIN CAPITAL LETTER E */
0x0046,     /* 46  LATIN CAPITAL LETTER F */
0x0047,     /* 47  LATIN CAPITAL LETTER G */
0x0048,     /* 48  LATIN CAPITAL LETTER H */
0x0049,     /* 49  LATIN CAPITAL LETTER I */
0x004A,     /* 4A  LATIN CAPITAL LETTER J */
0x004B,     /* 4B  LATIN CAPITAL LETTER K */
0x004C,     /* 4C  LATIN CAPITAL LETTER L */
0x004D,     /* 4D  LATIN CAPITAL LETTER M */
0x004E,     /* 4E  LATIN CAPITAL LETTER N */
0x004F,     /* 4F  LATIN CAPITAL LETTER O */
0x0050,     /* 50  LATIN CAPITAL LETTER P */
0x0051,     /* 51  LATIN CAPITAL LETTER Q */
0x0052,     /* 52  LATIN CAPITAL LETTER R */
0x0053,     /* 53  LATIN CAPITAL LETTER S */
0x0054,     /* 54  LATIN CAPITAL LETTER T */
0x0055,     /* 55  LATIN CAPITAL LETTER U */
0x0056,     /* 56  LATIN CAPITAL LETTER V */
0x0057,     /* 57  LATIN CAPITAL LETTER W */
0x0058,     /* 58  LATIN CAPITAL LETTER X */
0x0059,     /* 59  LATIN CAPITAL LETTER Y */
0x005A,     /* 5A  LATIN CAPITAL LETTER Z */
0x005B,     /* 5B  LEFT SQUARE BRACKET */
0x005C,     /* 5C  REVERSE SOLIDUS */
0x005D,     /* 5D  RIGHT SQUARE BRACKET */
0x005E,     /* 5E  CIRCUMFLEX ACCENT */
0x005F,     /* 5F  LOW LINE */
0x0060,     /* 60  GRAVE ACCENT */
0x0061,     /* 61  LATIN SMALL LETTER A */
0x0062,     /* 62  LATIN SMALL LETTER B */
0x0063,     /* 63  LATIN SMALL LETTER C */
0x0064,     /* 64  LATIN SMALL LETTER D */
0x0065,     /* 65  LATIN SMALL LETTER E */
0x0066,     /* 66  LATIN SMALL LETTER F */
0x0067,     /* 67  LATIN SMALL LETTER G */
0x0068,     /* 68  LATIN SMALL LETTER H */
0x0069,     /* 69  LATIN SMALL LETTER I */
0x006A,     /* 6A  LATIN SMALL LETTER J */
0x006B,     /* 6B  LATIN SMALL LETTER K */
0x006C,     /* 6C  LATIN SMALL LETTER L */
0x006D,     /* 6D  LATIN SMALL LETTER M */
0x006E,     /* 6E  LATIN SMALL LETTER N */
0x006F,     /* 6F  LATIN SMALL LETTER O */
0x0070,     /* 70  LATIN SMALL LETTER P */
0x0071,     /* 71  LATIN SMALL LETTER Q */
0x0072,     /* 72  LATIN SMALL LETTER R */
0x0073,     /* 73  LATIN SMALL LETTER S */
0x0074,     /* 74  LATIN SMALL LETTER T */
0x0075,     /* 75  LATIN SMALL LETTER U */
0x0076,     /* 76  LATIN SMALL LETTER V */
0x0077,     /* 77  LATIN SMALL LETTER W */
0x0078,     /* 78  LATIN SMALL LETTER X */
0x0079,     /* 79  LATIN SMALL LETTER Y */
0x007A,     /* 7A  LATIN SMALL LETTER Z */
0x007B,     /* 7B  LEFT CURLY BRACKET */
0x007C,     /* 7C  VERTICAL LINE */
0x007D,     /* 7D  RIGHT CURLY BRACKET */
0x007E,     /* 7E  TILDE */
UV_UNDEF,   /* 7F */
0x0410,     /* 80  CYRILLIC CAPITAL LETTER A */
0x0411,     /* 81  CYRILLIC CAPITAL LETTER BE */
0x0412,     /* 82  CYRILLIC CAPITAL LETTER VE */
0x0413,     /* 83  CYRILLIC CAPITAL LETTER GHE */
0x0414,     /* 84  CYRILLIC CAPITAL LETTER DE */
0x0415,     /* 85  CYRILLIC CAPITAL LETTER IE */
0x0416,     /* 86  CYRILLIC CAPITAL LETTER ZHE */
0x0417,     /* 87  CYRILLIC CAPITAL LETTER ZE */
0x0418,     /* 88  CYRILLIC CAPITAL LETTER I */
0x0419,     /* 89  CYRILLIC CAPITAL LETTER SHORT I */
0x041A,     /* 8A  CYRILLIC CAPITAL LETTER KA */
0x041B,     /* 8B  CYRILLIC CAPITAL LETTER EL */
0x041C,     /* 8C  CYRILLIC CAPITAL LETTER EM */
0x041D,     /* 8D  CYRILLIC CAPITAL LETTER EN */
0x041E,     /* 8E  CYRILLIC CAPITAL LETTER O */
0x041F,     /* 8F  CYRILLIC CAPITAL LETTER PE */
0x0420,     /* 90  CYRILLIC CAPITAL LETTER ER */
0x0421,     /* 91  CYRILLIC CAPITAL LETTER ES */
0x0422,     /* 92  CYRILLIC CAPITAL LETTER TE */
0x0423,     /* 93  CYRILLIC CAPITAL LETTER U */
0x0424,     /* 94  CYRILLIC CAPITAL LETTER EF */
0x0425,     /* 95  CYRILLIC CAPITAL LETTER HA */
0x0426,     /* 96  CYRILLIC CAPITAL LETTER TSE */
0x0427,     /* 97  CYRILLIC CAPITAL LETTER CHE */
0x0428,     /* 98  CYRILLIC CAPITAL LETTER SHA */
0x0429,     /* 99  CYRILLIC CAPITAL LETTER SHCHA */
0x042A,     /* 9A  CYRILLIC CAPITAL LETTER HARD SIGN */
0x042B,     /* 9B  CYRILLIC CAPITAL LETTER YERU */
0x042C,     /* 9C  CYRILLIC CAPITAL LETTER SOFT SIGN */
0x042D,     /* 9D  CYRILLIC CAPITAL LETTER E */
0x042E,     /* 9E  CYRILLIC CAPITAL LETTER YU */
0x042F,     /* 9F  CYRILLIC CAPITAL LETTER YA */
0x2020,     /* A0  DAGGER */
0x00B0,     /* A1  DEGREE SIGN */
0x0490,     /* A2  CYRILLIC CAPITAL LETTER GHE WITH UPTURN */
0x00A3,     /* A3  POUND SIGN */
0x00A7,     /* A4  SECTION SIGN */
0x2022,     /* A5  BULLET */
0x00B6,     /* A6  PILCROW SIGN */
0x0406,     /* A7  CYRILLIC CAPITAL LETTER BYELORUSSIAN-UKRAINIAN I */
0x00AE,     /* A8  REGISTERED SIGN */
0x00A9,     /* A9  COPYRIGHT SIGN */
0x2122,     /* AA  TRADE MARK SIGN */
0x0402,     /* AB  CYRILLIC CAPITAL LETTER DJE */
0x0452,     /* AC  CYRILLIC SMALL LETTER DJE */
0x2260,     /* AD  NOT EQUAL TO */
0x0403,     /* AE  CYRILLIC CAPITAL LETTER GJE */
0x0453,     /* AF  CYRILLIC SMALL LETTER GJE */
0x221E,     /* B0  INFINITY */
0x00B1,     /* B1  PLUS-MINUS SIGN */
0x2264,     /* B2  LESS-THAN OR EQUAL TO */
0x2265,     /* B3  GREATER-THAN OR EQUAL TO */
0x0456,     /* B4  CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I */
0x00B5,     /* B5  MICRO SIGN */
0x0491,     /* B6  CYRILLIC SMALL LETTER GHE WITH UPTURN */
0x0408,     /* B7  CYRILLIC CAPITAL LETTER JE */
0x0404,     /* B8  CYRILLIC CAPITAL LETTER UKRAINIAN IE */
0x0454,     /* B9  CYRILLIC SMALL LETTER UKRAINIAN IE */
0x0407,     /* BA  CYRILLIC CAPITAL LETTER YI */
0x0457,     /* BB  CYRILLIC SMALL LETTER YI */
0x0409,     /* BC  CYRILLIC CAPITAL LETTER LJE */
0x0459,     /* BD  CYRILLIC SMALL LETTER LJE */
0x040A,     /* BE  CYRILLIC CAPITAL LETTER NJE */
0x045A,     /* BF  CYRILLIC SMALL LETTER NJE */
0x0458,     /* C0  CYRILLIC SMALL LETTER JE */
0x0405,     /* C1  CYRILLIC CAPITAL LETTER DZE */
0x00AC,     /* C2  NOT SIGN */
0x221A,     /* C3  SQUARE ROOT */
0x0192,     /* C4  LATIN SMALL LETTER F WITH HOOK */
0x2248,     /* C5  ALMOST EQUAL TO */
0x2206,     /* C6  INCREMENT */
0x00AB,     /* C7  LEFT-POINTING DOUBLE ANGLE QUOTATION MARK */
0x00BB,     /* C8  RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK */
0x2026,     /* C9  HORIZONTAL ELLIPSIS */
0x00A0,     /* CA  NO-BREAK SPACE */
0x040B,     /* CB  CYRILLIC CAPITAL LETTER TSHE */
0x045B,     /* CC  CYRILLIC SMALL LETTER TSHE */
0x040C,     /* CD  CYRILLIC CAPITAL LETTER KJE */
0x045C,     /* CE  CYRILLIC SMALL LETTER KJE */
0x0455,     /* CF  CYRILLIC SMALL LETTER DZE */
0x2013,     /* D0  EN DASH */
0x2014,     /* D1  EM DASH */
0x201C,     /* D2  LEFT DOUBLE QUOTATION MARK */
0x201D,     /* D3  RIGHT DOUBLE QUOTATION MARK */
0x2018,     /* D4  LEFT SINGLE QUOTATION MARK */
0x2019,     /* D5  RIGHT SINGLE QUOTATION MARK */
0x00F7,     /* D6  DIVISION SIGN */
0x201E,     /* D7  DOUBLE LOW-9 QUOTATION MARK */
0x040E,     /* D8  CYRILLIC CAPITAL LETTER SHORT U */
0x045E,     /* D9  CYRILLIC SMALL LETTER SHORT U */
0x040F,     /* DA  CYRILLIC CAPITAL LETTER DZHE */
0x045F,     /* DB  CYRILLIC SMALL LETTER DZHE */
0x2116,     /* DC  NUMERO SIGN */
0x0401,     /* DD  CYRILLIC CAPITAL LETTER IO */
0x0451,     /* DE  CYRILLIC SMALL LETTER IO */
0x044F,     /* DF  CYRILLIC SMALL LETTER YA */
0x0430,     /* E0  CYRILLIC SMALL LETTER A */
0x0431,     /* E1  CYRILLIC SMALL LETTER BE */
0x0432,     /* E2  CYRILLIC SMALL LETTER VE */
0x0433,     /* E3  CYRILLIC SMALL LETTER GHE */
0x0434,     /* E4  CYRILLIC SMALL LETTER DE */
0x0435,     /* E5  CYRILLIC SMALL LETTER IE */
0x0436,     /* E6  CYRILLIC SMALL LETTER ZHE */
0x0437,     /* E7  CYRILLIC SMALL LETTER ZE */
0x0438,     /* E8  CYRILLIC SMALL LETTER I */
0x0439,     /* E9  CYRILLIC SMALL LETTER SHORT I */
0x043A,     /* EA  CYRILLIC SMALL LETTER KA */
0x043B,     /* EB  CYRILLIC SMALL LETTER EL */
0x043C,     /* EC  CYRILLIC SMALL LETTER EM */
0x043D,     /* ED  CYRILLIC SMALL LETTER EN */
0x043E,     /* EE  CYRILLIC SMALL LETTER O */
0x043F,     /* EF  CYRILLIC SMALL LETTER PE */
0x0440,     /* F0  CYRILLIC SMALL LETTER ER */
0x0441,     /* F1  CYRILLIC SMALL LETTER ES */
0x0442,     /* F2  CYRILLIC SMALL LETTER TE */
0x0443,     /* F3  CYRILLIC SMALL LETTER U */
0x0444,     /* F4  CYRILLIC SMALL LETTER EF */
0x0445,     /* F5  CYRILLIC SMALL LETTER HA */
0x0446,     /* F6  CYRILLIC SMALL LETTER TSE */
0x0447,     /* F7  CYRILLIC SMALL LETTER CHE */
0x0448,     /* F8  CYRILLIC SMALL LETTER SHA */
0x0449,     /* F9  CYRILLIC SMALL LETTER SHCHA */
0x044A,     /* FA  CYRILLIC SMALL LETTER HARD SIGN */
0x044B,     /* FB  CYRILLIC SMALL LETTER YERU */
0x044C,     /* FC  CYRILLIC SMALL LETTER SOFT SIGN */
0x044D,     /* FD  CYRILLIC SMALL LETTER E */
0x044E,     /* FE  CYRILLIC SMALL LETTER YU */
0x20AC,     /* FF  EURO SIGN */
