"[-svg options: defaults -lf, -gn0]\n"
"-lf       Unix newline [default]\n"
"-cr       Mac newline\n"
"-crlf     Windows newline\n"
"\n"
"-gn0      Glyph names for no glyphs [default]\n"
"-gn1      Glyph names for non-ASCII glyphs only\n"
"-gn2      Glyph names for all glyphs\n",
"\n"
"-abs      Write absolute coordinates (rather than relative)\n"
"-sa       Standalone font file\n"
"\n"
"SVG mode converts an abstract font to an SVG font. The form of the SVG font is\n"
"controlled by the options above.\n"
"\n"
"The options are reasonably self explanatory.\n"
"\n",
"For example, the command:\n"
"\n"
"    tx -svg -crlf -gn1 rdr_____.pfb\n"
"\n"
"converts the font file \"rdr_____.pfb\" to an SVG font with Windows newlines and\n"
"glyph names for non-ASCII Unicode characters, and writes the result to stdout.\n"
"\n"
"The command:\n",
"\n"
"    tx -svg -sa rdr_____.pfb test.svg\n"
"\n"
"converts the font file \"rdr_____.pfb\" to a standalone SVG font with Unix\n"
"newlines and no glyph names, and writes the result to test.svg. The result is\n"
"a self-contained SVG font file.\n"
