/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
   This software is licensed as OpenSource, under the Apache License, Version 2.0. This license is available at: http://opensource.org/licenses/Apache-2.0. */
/***********************************************************************/

/*
 * Presence of at least half (13) of the following UVs constitutes support of
 * Simplified (as opposed to Traditional) Chinese, given that the CJK Unified
 * Ideographs and Bopomofo Unicode ranges are supported.
 */

0x00E0,     /* LATIN SMALL LETTER A WITH GRAVE */
    0x00E1, /* LATIN SMALL LETTER A WITH ACUTE */
    0x00E8, /* LATIN SMALL LETTER E WITH GRAVE */
    0x00E9, /* LATIN SMALL LETTER E WITH ACUTE */
    0x00EA, /* LATIN SMALL LETTER E WITH CIRCUMFLEX */
    0x00EC, /* LATIN SMALL LETTER I WITH GRAVE */
    0x00ED, /* LATIN SMALL LETTER I WITH ACUTE */
    0x00F2, /* LATIN SMALL LETTER O WITH GRAVE */
    0x00F3, /* LATIN SMALL LETTER O WITH ACUTE */
    0x00F9, /* LATIN SMALL LETTER U WITH GRAVE */
    0x00FA, /* LATIN SMALL LETTER U WITH ACUTE */
    0x00FC, /* LATIN SMALL LETTER U WITH DIAERESIS */
    0x0101, /* LATIN SMALL LETTER A WITH MACRON */
    0x0113, /* LATIN SMALL LETTER E WITH MACRON */
    0x011B, /* LATIN SMALL LETTER E WITH CARON */
    0x012B, /* LATIN SMALL LETTER I WITH MACRON */
    0x014D, /* LATIN SMALL LETTER O WITH MACRON */
    0x016B, /* LATIN SMALL LETTER U WITH MACRON */
    0x01CE, /* LATIN SMALL LETTER A WITH CARON */
    0x01D0, /* LATIN SMALL LETTER I WITH CARON */
    0x01D2, /* LATIN SMALL LETTER O WITH CARON */
    0x01D4, /* LATIN SMALL LETTER U WITH CARON */
    0x01D6, /* LATIN SMALL LETTER U WITH DIAERESIS AND MACRON */
    0x01D8, /* LATIN SMALL LETTER U WITH DIAERESIS AND ACUTE */
    0x01DA, /* LATIN SMALL LETTER U WITH DIAERESIS AND CARON */
    0x01DC, /* LATIN SMALL LETTER U WITH DIAERESIS AND GRAVE */
