/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
   This software is licensed as OpenSource, under the Apache License, Version 2.0. This license is available at: http://opensource.org/licenses/Apache-2.0. */
/***********************************************************************/

#ifndef XXXX_H
#define XXXX_H

#include "common.h"

#define xxxx_ TAG('x', 'x', 'x', 'x')

/* Standard functions */
void xxxxNew(hotCtx g);
int xxxxFill(hotCtx g);
void xxxxWrite(hotCtx g);
void xxxxReuse(hotCtx g);
void xxxxFree(hotCtx g);

/* Supplementary functions */

#endif /* XXXX_H */
