/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
   This software is licensed as OpenSource, under the Apache License, Version 2.0.
   This license is available at: http://opensource.org/licenses/Apache-2.0. */

/* Mac OS Farsi aggregate Unicode initializer.

   Element values are UVs. UV_UNDEF is 0xFFFF. Index by code, get UV.
   Source: ftp://ftp.unicode.org/Public/MAPPINGS/VENDORS/APPLE/FARSI.TXT
   as of 9/9/99. */

UV_UNDEF,   /* 00 */
UV_UNDEF,   /* 01 */
UV_UNDEF,   /* 02 */
UV_UNDEF,   /* 03 */
UV_UNDEF,   /* 04 */
UV_UNDEF,   /* 05 */
UV_UNDEF,   /* 06 */
UV_UNDEF,   /* 07 */
UV_UNDEF,   /* 08 */
UV_UNDEF,   /* 09 */
UV_UNDEF,   /* 0A */
UV_UNDEF,   /* 0B */
UV_UNDEF,   /* 0C */
UV_UNDEF,   /* 0D */
UV_UNDEF,   /* 0E */
UV_UNDEF,   /* 0F */
UV_UNDEF,   /* 10 */
UV_UNDEF,   /* 11 */
UV_UNDEF,   /* 12 */
UV_UNDEF,   /* 13 */
UV_UNDEF,   /* 14 */
UV_UNDEF,   /* 15 */
UV_UNDEF,   /* 16 */
UV_UNDEF,   /* 17 */
UV_UNDEF,   /* 18 */
UV_UNDEF,   /* 19 */
UV_UNDEF,   /* 1A */
UV_UNDEF,   /* 1B */
UV_UNDEF,   /* 1C */
UV_UNDEF,   /* 1D */
UV_UNDEF,   /* 1E */
UV_UNDEF,   /* 1F */
0x0020,     /* 20  SPACE, left-right */
0x0021,     /* 21  EXCLAMATION MARK, left-right */
0x0022,     /* 22  QUOTATION MARK, left-right */
0x0023,     /* 23  NUMBER SIGN, left-right */
0x0024,     /* 24  DOLLAR SIGN, left-right */
0x0025,     /* 25  PERCENT SIGN, left-right */
0x0026,     /* 26  AMPERSAND, left-right */
0x0027,     /* 27  APOSTROPHE, left-right */
0x0028,     /* 28  LEFT PARENTHESIS, left-right */
0x0029,     /* 29  RIGHT PARENTHESIS, left-right */
0x002A,     /* 2A  ASTERISK, left-right */
0x002B,     /* 2B  PLUS SIGN, left-right */
0x002C,     /* 2C  COMMA, left-right */
0x002D,     /* 2D  HYPHEN-MINUS, left-right */
0x002E,     /* 2E  FULL STOP, left-right */
0x002F,     /* 2F  SOLIDUS, left-right */
0x0030,     /* 30  DIGIT ZERO */
0x0031,     /* 31  DIGIT ONE */
0x0032,     /* 32  DIGIT TWO */
0x0033,     /* 33  DIGIT THREE */
0x0034,     /* 34  DIGIT FOUR */
0x0035,     /* 35  DIGIT FIVE */
0x0036,     /* 36  DIGIT SIX */
0x0037,     /* 37  DIGIT SEVEN */
0x0038,     /* 38  DIGIT EIGHT */
0x0039,     /* 39  DIGIT NINE */
0x003A,     /* 3A  COLON, left-right */
0x003B,     /* 3B  SEMICOLON, left-right */
0x003C,     /* 3C  LESS-THAN SIGN, left-right */
0x003D,     /* 3D  EQUALS SIGN, left-right */
0x003E,     /* 3E  GREATER-THAN SIGN, left-right */
0x003F,     /* 3F  QUESTION MARK, left-right */
0x0040,     /* 40  COMMERCIAL AT */
0x0041,     /* 41  LATIN CAPITAL LETTER A */
0x0042,     /* 42  LATIN CAPITAL LETTER B */
0x0043,     /* 43  LATIN CAPITAL LETTER C */
0x0044,     /* 44  LATIN CAPITAL LETTER D */
0x0045,     /* 45  LATIN CAPITAL LETTER E */
0x0046,     /* 46  LATIN CAPITAL LETTER F */
0x0047,     /* 47  LATIN CAPITAL LETTER G */
0x0048,     /* 48  LATIN CAPITAL LETTER H */
0x0049,     /* 49  LATIN CAPITAL LETTER I */
0x004A,     /* 4A  LATIN CAPITAL LETTER J */
0x004B,     /* 4B  LATIN CAPITAL LETTER K */
0x004C,     /* 4C  LATIN CAPITAL LETTER L */
0x004D,     /* 4D  LATIN CAPITAL LETTER M */
0x004E,     /* 4E  LATIN CAPITAL LETTER N */
0x004F,     /* 4F  LATIN CAPITAL LETTER O */
0x0050,     /* 50  LATIN CAPITAL LETTER P */
0x0051,     /* 51  LATIN CAPITAL LETTER Q */
0x0052,     /* 52  LATIN CAPITAL LETTER R */
0x0053,     /* 53  LATIN CAPITAL LETTER S */
0x0054,     /* 54  LATIN CAPITAL LETTER T */
0x0055,     /* 55  LATIN CAPITAL LETTER U */
0x0056,     /* 56  LATIN CAPITAL LETTER V */
0x0057,     /* 57  LATIN CAPITAL LETTER W */
0x0058,     /* 58  LATIN CAPITAL LETTER X */
0x0059,     /* 59  LATIN CAPITAL LETTER Y */
0x005A,     /* 5A  LATIN CAPITAL LETTER Z */
0x005B,     /* 5B  LEFT SQUARE BRACKET, left-right */
0x005C,     /* 5C  REVERSE SOLIDUS, left-right */
0x005D,     /* 5D  RIGHT SQUARE BRACKET, left-right */
0x005E,     /* 5E  CIRCUMFLEX ACCENT, left-right */
0x005F,     /* 5F  LOW LINE, left-right */
0x0060,     /* 60  GRAVE ACCENT */
0x0061,     /* 61  LATIN SMALL LETTER A */
0x0062,     /* 62  LATIN SMALL LETTER B */
0x0063,     /* 63  LATIN SMALL LETTER C */
0x0064,     /* 64  LATIN SMALL LETTER D */
0x0065,     /* 65  LATIN SMALL LETTER E */
0x0066,     /* 66  LATIN SMALL LETTER F */
0x0067,     /* 67  LATIN SMALL LETTER G */
0x0068,     /* 68  LATIN SMALL LETTER H */
0x0069,     /* 69  LATIN SMALL LETTER I */
0x006A,     /* 6A  LATIN SMALL LETTER J */
0x006B,     /* 6B  LATIN SMALL LETTER K */
0x006C,     /* 6C  LATIN SMALL LETTER L */
0x006D,     /* 6D  LATIN SMALL LETTER M */
0x006E,     /* 6E  LATIN SMALL LETTER N */
0x006F,     /* 6F  LATIN SMALL LETTER O */
0x0070,     /* 70  LATIN SMALL LETTER P */
0x0071,     /* 71  LATIN SMALL LETTER Q */
0x0072,     /* 72  LATIN SMALL LETTER R */
0x0073,     /* 73  LATIN SMALL LETTER S */
0x0074,     /* 74  LATIN SMALL LETTER T */
0x0075,     /* 75  LATIN SMALL LETTER U */
0x0076,     /* 76  LATIN SMALL LETTER V */
0x0077,     /* 77  LATIN SMALL LETTER W */
0x0078,     /* 78  LATIN SMALL LETTER X */
0x0079,     /* 79  LATIN SMALL LETTER Y */
0x007A,     /* 7A  LATIN SMALL LETTER Z */
0x007B,     /* 7B  LEFT CURLY BRACKET, left-right */
0x007C,     /* 7C  VERTICAL LINE, left-right */
0x007D,     /* 7D  RIGHT CURLY BRACKET, left-right */
0x007E,     /* 7E  TILDE */
UV_UNDEF,   /* 7F */
0x00C4,     /* 80  LATIN CAPITAL LETTER A WITH DIAERESIS */
0x00A0,     /* 81  NO-BREAK SPACE, right-left */
0x00C7,     /* 82  LATIN CAPITAL LETTER C WITH CEDILLA */
0x00C9,     /* 83  LATIN CAPITAL LETTER E WITH ACUTE */
0x00D1,     /* 84  LATIN CAPITAL LETTER N WITH TILDE */
0x00D6,     /* 85  LATIN CAPITAL LETTER O WITH DIAERESIS */
0x00DC,     /* 86  LATIN CAPITAL LETTER U WITH DIAERESIS */
0x00E1,     /* 87  LATIN SMALL LETTER A WITH ACUTE */
0x00E0,     /* 88  LATIN SMALL LETTER A WITH GRAVE */
0x00E2,     /* 89  LATIN SMALL LETTER A WITH CIRCUMFLEX */
0x00E4,     /* 8A  LATIN SMALL LETTER A WITH DIAERESIS */
0x06BA,     /* 8B  ARABIC LETTER NOON GHUNNA */
0x00AB,     /* 8C  LEFT-POINTING DOUBLE ANGLE QUOTATION MARK, right-left */
0x00E7,     /* 8D  LATIN SMALL LETTER C WITH CEDILLA */
0x00E9,     /* 8E  LATIN SMALL LETTER E WITH ACUTE */
0x00E8,     /* 8F  LATIN SMALL LETTER E WITH GRAVE */
0x00EA,     /* 90  LATIN SMALL LETTER E WITH CIRCUMFLEX */
0x00EB,     /* 91  LATIN SMALL LETTER E WITH DIAERESIS */
0x00ED,     /* 92  LATIN SMALL LETTER I WITH ACUTE */
0x2026,     /* 93  HORIZONTAL ELLIPSIS, right-left */
0x00EE,     /* 94  LATIN SMALL LETTER I WITH CIRCUMFLEX */
0x00EF,     /* 95  LATIN SMALL LETTER I WITH DIAERESIS */
0x00F1,     /* 96  LATIN SMALL LETTER N WITH TILDE */
0x00F3,     /* 97  LATIN SMALL LETTER O WITH ACUTE */
0x00BB,     /* 98  RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK, right-left */
0x00F4,     /* 99  LATIN SMALL LETTER O WITH CIRCUMFLEX */
0x00F6,     /* 9A  LATIN SMALL LETTER O WITH DIAERESIS */
0x00F7,     /* 9B  DIVISION SIGN, right-left */
0x00FA,     /* 9C  LATIN SMALL LETTER U WITH ACUTE */
0x00F9,     /* 9D  LATIN SMALL LETTER U WITH GRAVE */
0x00FB,     /* 9E  LATIN SMALL LETTER U WITH CIRCUMFLEX */
0x00FC,     /* 9F  LATIN SMALL LETTER U WITH DIAERESIS */
0x0020,     /* A0  SPACE, right-left */
0x0021,     /* A1  EXCLAMATION MARK, right-left */
0x0022,     /* A2  QUOTATION MARK, right-left */
0x0023,     /* A3  NUMBER SIGN, right-left */
0x0024,     /* A4  DOLLAR SIGN, right-left */
0x066A,     /* A5  ARABIC PERCENT SIGN */
0x0026,     /* A6  AMPERSAND, right-left */
0x0027,     /* A7  APOSTROPHE, right-left */
0x0028,     /* A8  LEFT PARENTHESIS, right-left */
0x0029,     /* A9  RIGHT PARENTHESIS, right-left */
0x002A,     /* AA  ASTERISK, right-left */
0x002B,     /* AB  PLUS SIGN, right-left */
0x060C,     /* AC  ARABIC COMMA */
0x002D,     /* AD  HYPHEN-MINUS, right-left */
0x002E,     /* AE  FULL STOP, right-left */
0x002F,     /* AF  SOLIDUS, right-left */
0x06F0,     /* B0  EXTENDED ARABIC-INDIC DIGIT ZERO, right-left */
0x06F1,     /* B1  EXTENDED ARABIC-INDIC DIGIT ONE, right-left */
0x06F2,     /* B2  EXTENDED ARABIC-INDIC DIGIT TWO, right-left */
0x06F3,     /* B3  EXTENDED ARABIC-INDIC DIGIT THREE, right-left */
0x06F4,     /* B4  EXTENDED ARABIC-INDIC DIGIT FOUR, right-left */
0x06F5,     /* B5  EXTENDED ARABIC-INDIC DIGIT FIVE, right-left */
0x06F6,     /* B6  EXTENDED ARABIC-INDIC DIGIT SIX, right-left */
0x06F7,     /* B7  EXTENDED ARABIC-INDIC DIGIT SEVEN, right-left */
0x06F8,     /* B8  EXTENDED ARABIC-INDIC DIGIT EIGHT, right-left */
0x06F9,     /* B9  EXTENDED ARABIC-INDIC DIGIT NINE, right-left */
0x003A,     /* BA  COLON, right-left */
0x061B,     /* BB  ARABIC SEMICOLON */
0x003C,     /* BC  LESS-THAN SIGN, right-left */
0x003D,     /* BD  EQUALS SIGN, right-left */
0x003E,     /* BE  GREATER-THAN SIGN, right-left */
0x061F,     /* BF  ARABIC QUESTION MARK */
0x274A,     /* C0  EIGHT TEARDROP-SPOKED PROPELLER ASTERISK, right-left */
0x0621,     /* C1  ARABIC LETTER HAMZA */
0x0622,     /* C2  ARABIC LETTER ALEF WITH MADDA ABOVE */
0x0623,     /* C3  ARABIC LETTER ALEF WITH HAMZA ABOVE */
0x0624,     /* C4  ARABIC LETTER WAW WITH HAMZA ABOVE */
0x0625,     /* C5  ARABIC LETTER ALEF WITH HAMZA BELOW */
0x0626,     /* C6  ARABIC LETTER YEH WITH HAMZA ABOVE */
0x0627,     /* C7  ARABIC LETTER ALEF */
0x0628,     /* C8  ARABIC LETTER BEH */
0x0629,     /* C9  ARABIC LETTER TEH MARBUTA */
0x062A,     /* CA  ARABIC LETTER TEH */
0x062B,     /* CB  ARABIC LETTER THEH */
0x062C,     /* CC  ARABIC LETTER JEEM */
0x062D,     /* CD  ARABIC LETTER HAH */
0x062E,     /* CE  ARABIC LETTER KHAH */
0x062F,     /* CF  ARABIC LETTER DAL */
0x0630,     /* D0  ARABIC LETTER THAL */
0x0631,     /* D1  ARABIC LETTER REH */
0x0632,     /* D2  ARABIC LETTER ZAIN */
0x0633,     /* D3  ARABIC LETTER SEEN */
0x0634,     /* D4  ARABIC LETTER SHEEN */
0x0635,     /* D5  ARABIC LETTER SAD */
0x0636,     /* D6  ARABIC LETTER DAD */
0x0637,     /* D7  ARABIC LETTER TAH */
0x0638,     /* D8  ARABIC LETTER ZAH */
0x0639,     /* D9  ARABIC LETTER AIN */
0x063A,     /* DA  ARABIC LETTER GHAIN */
0x005B,     /* DB  LEFT SQUARE BRACKET, right-left */
0x005C,     /* DC  REVERSE SOLIDUS, right-left */
0x005D,     /* DD  RIGHT SQUARE BRACKET, right-left */
0x005E,     /* DE  CIRCUMFLEX ACCENT, right-left */
0x005F,     /* DF  LOW LINE, right-left */
0x0640,     /* E0  ARABIC TATWEEL */
0x0641,     /* E1  ARABIC LETTER FEH */
0x0642,     /* E2  ARABIC LETTER QAF */
0x0643,     /* E3  ARABIC LETTER KAF */
0x0644,     /* E4  ARABIC LETTER LAM */
0x0645,     /* E5  ARABIC LETTER MEEM */
0x0646,     /* E6  ARABIC LETTER NOON */
0x0647,     /* E7  ARABIC LETTER HEH */
0x0648,     /* E8  ARABIC LETTER WAW */
0x0649,     /* E9  ARABIC LETTER ALEF MAKSURA */
0x064A,     /* EA  ARABIC LETTER YEH */
0x064B,     /* EB  ARABIC FATHATAN */
0x064C,     /* EC  ARABIC DAMMATAN */
0x064D,     /* ED  ARABIC KASRATAN */
0x064E,     /* EE  ARABIC FATHA */
0x064F,     /* EF  ARABIC DAMMA */
0x0650,     /* F0  ARABIC KASRA */
0x0651,     /* F1  ARABIC SHADDA */
0x0652,     /* F2  ARABIC SUKUN */
0x067E,     /* F3  ARABIC LETTER PEH */
0x0679,     /* F4  ARABIC LETTER TTEH */
0x0686,     /* F5  ARABIC LETTER TCHEH */
0x06D5,     /* F6  ARABIC LETTER AE */
0x06A4,     /* F7  ARABIC LETTER VEH */
0x06AF,     /* F8  ARABIC LETTER GAF */
0x0688,     /* F9  ARABIC LETTER DDAL */
0x0691,     /* FA  ARABIC LETTER RREH */
0x007B,     /* FB  LEFT CURLY BRACKET, right-left */
0x007C,     /* FC  VERTICAL LINE, right-left */
0x007D,     /* FD  RIGHT CURLY BRACKET, right-left */
0x0698,     /* FE  ARABIC LETTER JEH */
0x06D2,     /* FF  ARABIC LETTER YEH BARREE */
