/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
   This software is licensed as OpenSource, under the Apache License, Version 2.0.
   This license is available at: http://opensource.org/licenses/Apache-2.0. */

#ifndef PACKAGE_SPECS_H
#define PACKAGE_SPECS_H

#define CFFREAD     "cffread.h"
#define DICTOPS     "dictops.h"
#define EXCS0       "excs0.h"
#define EXENC0      "exenc0.h"
#define EXSUBCS0    "exsubcs0.h"
#define ISOCS0      "isocs0.h"
#define STDSTR1     "stdstr1.h"
#define TXOPS       "txops.h"

#endif /* PACKAGE_SPECS_H */
