/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
   This software is licensed as OpenSource, under the Apache License, Version 2.0.
   This license is available at: http://opensource.org/licenses/Apache-2.0. */

/* Must be alphabetically sorted */
DCL_KEY('B', k_cmp)
DCL_KEY('C', k_ct)
DCL_KEY('L', k_dt)
DCL_KEY('M', k_mt)
DCL_KEY('R', k_trans)
DCL_KEY('Z', k_CP)
DCL_KEY('c', k_rct)
DCL_KEY('l', k_rdt)
DCL_KEY('m', k_rmt)
DCL_KEY('z', k_cp)
