{"afii10017", "A.cyr"},
{"afii10018", "Be"},
{"afii10019", "Ve"},
{"afii10020", "Ghe"},
{"afii10021", "De"},
{"afii10022", "Ie"},
{"afii10023", "Io"},
{"afii10024", "Zhe"},
{"afii10025", "Ze"},
{"afii10026", "I.cyr"},
{"afii10027", "I.short"},
{"afii10028", "Ka"},
{"afii10029", "El"},
{"afii10030", "Em"},
{"afii10031", "En"},
{"afii10032", "O.cyr"},
{"afii10033", "Pe"},
{"afii10034", "Er"},
{"afii10035", "Es"},
{"afii10036", "Te"},
{"afii10037", "U.cyr"},
{"afii10038", "Ef"},
{"afii10039", "Ha"},
{"afii10040", "Tse"},
{"afii10041", "Che"},
{"afii10042", "Sha"},
{"afii10043", "Shcha"},
{"afii10044", "Hard"},
{"afii10045", "Yeru"},
{"afii10046", "Soft"},
{"afii10047", "E.cyr"},
{"afii10048", "Yu"},
{"afii10049", "Ya"},
{"afii10050", "Ghe.up"},
{"afii10051", "Dje"},
{"afii10052", "Gje"},
{"afii10053", "Ie.ukran"},
{"afii10054", "Dze"},
{"afii10055", "I.ukran"},
{"afii10056", "Yi"},
{"afii10057", "Je"},
{"afii10058", "Lje"},
{"afii10059", "Nje"},
{"afii10060", "Tshe"},
{"afii10061", "Kje"},
{"afii10062", "U.short"},
{"afii10063", "ghe.ital"},
{"afii10064", "be.alt"},
{"afii10065", "a.cyr"},
{"afii10066", "be"},
{"afii10067", "ve"},
{"afii10068", "ghe"},
{"afii10069", "de"},
{"afii10070", "ie"},
{"afii10071", "io"},
{"afii10072", "zhe"},
{"afii10073", "ze"},
{"afii10074", "i.cyr"},
{"afii10075", "i.short"},
{"afii10076", "ka"},
{"afii10077", "el"},
{"afii10078", "em"},
{"afii10079", "en"},
{"afii10080", "o.cyr"},
{"afii10081", "pe"},
{"afii10082", "er"},
{"afii10083", "es"},
{"afii10084", "te"},
{"afii10085", "u.cyr"},
{"afii10086", "ef"},
{"afii10087", "ha"},
{"afii10088", "tse"},
{"afii10089", "che"},
{"afii10090", "sha"},
{"afii10091", "shcha"},
{"afii10092", "hard"},
{"afii10093", "yeru"},
{"afii10094", "soft"},
{"afii10095", "e.cyr"},
{"afii10096", "yu"},
{"afii10097", "ya"},
{"afii10098", "ghe.up"},
{"afii10099", "dje"},
{"afii10100", "gje"},
{"afii10101", "ie.ukran"},
{"afii10102", "dze"},
{"afii10103", "i.ukran"},
{"afii10104", "yi"},
{"afii10105", "je"},
{"afii10106", "lje"},
{"afii10107", "nje"},
{"afii10108", "tshe"},
{"afii10109", "kje"},
{"afii10110", "u.short"},
{"afii10145", "Dzhe"},
{"afii10146", "Yat"},
{"afii10147", "Fita"},
{"afii10148", "Izhitsa"},
{"afii10192", "de.ital"},
{"afii10193", "dzhe"},
{"afii10194", "yat"},
{"afii10195", "fita"},
{"afii10196", "izhitsa"},
{"afii10831", "pe.ital"},
{"afii10832", "te.ital"},
{"afii10846", "schwa"},
{"afii61248", "careof"},
{"afii61289", "litre"},
{"afii61352", "numero"},
{"uni007F", "control"},
{"uni00A0", "nbspace"},
{"uni00AD", "softhyphen"},
{"uni0374", "numeralsign"},
{"uni0375", "lownumeralsign"},
{"uni037A", "lowcomma"},
{"uni037E", "questiongreek"},
{"uni03BC", "mu.greek"},
{"uni03D0", "beta.alt"},
{"uni03D1", "theta.alt"},
{"uni03D2", "Upsilon.alt"},
{"uni03D5", "phi.alt"},
{"uni03D6", "pi.alt"},
{"uni03DA", "stigma"},
{"uni03DE", "koppa"},
{"uni03E0", "sampi"},
{"uni03F0", "kappa.alt"},
{"uni1F00", "alphalenis"},
{"uni1F01", "alphaasper"},
{"uni1F02", "alphalenisgrave"},
{"uni1F03", "alphaaspergrave"},
{"uni1F04", "alphalenisacute"},
{"uni1F05", "alphaasperacute"},
{"uni1F06", "alphalenistilde"},
{"uni1F07", "alphaaspertilde"},
{"uni1F08", "Alphalenis"},
{"uni1F09", "Alphaasper"},
{"uni1F0A", "Alphalenisgrave"},
{"uni1F0B", "Alphaaspergrave"},
{"uni1F0C", "Alphalenisacute"},
{"uni1F0D", "Alphaasperacute"},
{"uni1F0E", "Alphalenistilde"},
{"uni1F0F", "Alphaaspertilde"},
{"uni1F10", "epsilonlenis"},
{"uni1F11", "epsilonasper"},
{"uni1F12", "epsilonlenisgrave"},
{"uni1F13", "epsilonaspergrave"},
{"uni1F14", "epsilonlenisacute"},
{"uni1F15", "epsilonasperacute"},
{"uni1F18", "Epsilonlenis"},
{"uni1F19", "Epsilonasper"},
{"uni1F1A", "Epsilonlenisgrave"},
{"uni1F1B", "Epsilonaspergrave"},
{"uni1F1C", "Epsilonlenisacute"},
{"uni1F1D", "Epsilonasperacute"},
{"uni1F20", "etalenis"},
{"uni1F21", "etaasper"},
{"uni1F22", "etalenisgrave"},
{"uni1F23", "etaaspergrave"},
{"uni1F24", "etalenisacute"},
{"uni1F25", "etaasperacute"},
{"uni1F26", "etalenistilde"},
{"uni1F27", "etaaspertilde"},
{"uni1F28", "Etalenis"},
{"uni1F29", "Etaasper"},
{"uni1F2A", "Etalenisgrave"},
{"uni1F2B", "Etaaspergrave"},
{"uni1F2C", "Etalenisacute"},
{"uni1F2D", "Etaasperacute"},
{"uni1F2E", "Etalenistilde"},
{"uni1F2F", "Etaaspertilde"},
{"uni1F30", "iotalenis"},
{"uni1F31", "iotaasper"},
{"uni1F32", "iotalenisgrave"},
{"uni1F33", "iotaaspergrave"},
{"uni1F34", "iotalenisacute"},
{"uni1F35", "iotaasperacute"},
{"uni1F36", "iotalenistilde"},
{"uni1F37", "iotaaspertilde"},
{"uni1F38", "Iotalenis"},
{"uni1F39", "Iotaasper"},
{"uni1F3A", "Iotalenisgrave"},
{"uni1F3B", "Iotaaspergrave"},
{"uni1F3C", "Iotalenisacute"},
{"uni1F3D", "Iotaasperacute"},
{"uni1F3E", "Iotalenistilde"},
{"uni1F3F", "Iotaaspertilde"},
{"uni1F40", "omicronlenis"},
{"uni1F41", "omicronasper"},
{"uni1F42", "omicronlenisgrave"},
{"uni1F43", "omicronaspergrave"},
{"uni1F44", "omicronlenisacute"},
{"uni1F45", "omicronasperacute"},
{"uni1F48", "Omicronlenis"},
{"uni1F49", "Omicronasper"},
{"uni1F4A", "Omicronlenisgrave"},
{"uni1F4B", "Omicronaspergrave"},
{"uni1F4C", "Omicronlenisacute"},
{"uni1F4D", "Omicronasperacute"},
{"uni1F50", "upsilonlenis"},
{"uni1F51", "upsilonasper"},
{"uni1F52", "upsilonlenisgrave"},
{"uni1F53", "upsilonaspergrave"},
{"uni1F54", "upsilonlenisacute"},
{"uni1F55", "upsilonasperacute"},
{"uni1F56", "upsilonlenistilde"},
{"uni1F57", "upsilonaspertilde"},
{"uni1F59", "Upsilonasper"},
{"uni1F5B", "Upsilonaspergrave"},
{"uni1F5D", "Upsilonasperacute"},
{"uni1F5F", "Upsilonaspertilde"},
{"uni1F60", "omegalenis"},
{"uni1F61", "omegaasper"},
{"uni1F62", "omegalenisgrave"},
{"uni1F63", "omegaaspergrave"},
{"uni1F64", "omegalenisacute"},
{"uni1F65", "omegaasperacute"},
{"uni1F66", "omegalenistilde"},
{"uni1F67", "omegaaspertilde"},
{"uni1F68", "Omegalenis"},
{"uni1F69", "Omegaasper"},
{"uni1F6A", "Omegalenisgrave"},
{"uni1F6B", "Omegaaspergrave"},
{"uni1F6C", "Omegalenisacute"},
{"uni1F6D", "Omegaasperacute"},
{"uni1F6E", "Omegalenistilde"},
{"uni1F6F", "Omegaaspertilde"},
{"uni1F70", "alphagrave"},
{"uni1F71", "alphaacute"},
{"uni1F72", "epsilongrave"},
{"uni1F73", "epsilonacute"},
{"uni1F74", "etagrave"},
{"uni1F75", "etaacute"},
{"uni1F76", "iotagrave"},
{"uni1F77", "iotaacute"},
{"uni1F78", "omicrongrave"},
{"uni1F79", "omicronacute"},
{"uni1F7A", "upsilongrave"},
{"uni1F7B", "upsilonacute"},
{"uni1F7C", "omegagrave"},
{"uni1F7D", "omegaacute"},
{"uni1F80", "alphaiotasublenis"},
{"uni1F81", "alphaiotasubasper"},
{"uni1F82", "alphaiotasublenisgrave"},
{"uni1F83", "alphaiotasubaspergrave"},
{"uni1F84", "alphaiotasublenisacute"},
{"uni1F85", "alphaiotasubasperacute"},
{"uni1F86", "alphaiotasublenistilde"},
{"uni1F87", "alphaiotasubaspertilde"},
{"uni1F88", "Alphaiotasublenis"},
{"uni1F89", "Alphaiotasubasper"},
{"uni1F8A", "Alphaiotasublenisgrave"},
{"uni1F8B", "Alphaiotasubaspergrave"},
{"uni1F8C", "Alphaiotasublenisacute"},
{"uni1F8D", "Alphaiotasubasperacute"},
{"uni1F8E", "Alphaiotasublenistilde"},
{"uni1F8F", "Alphaiotasubaspertilde"},
{"uni1F90", "etaiotasublenis"},
{"uni1F91", "etaiotasubasper"},
{"uni1F92", "etaiotasublenisgrave"},
{"uni1F93", "etaiotasubaspergrave"},
{"uni1F94", "etaiotasublenisacute"},
{"uni1F95", "etaiotasubasperacute"},
{"uni1F96", "etaiotasublenistilde"},
{"uni1F97", "etaiotasubaspertilde"},
{"uni1F98", "Etaiotasublenis"},
{"uni1F99", "Etaiotasubasper"},
{"uni1F9A", "Etaiotasublenisgrave"},
{"uni1F9B", "Etaiotasubaspergrave"},
{"uni1F9C", "Etaiotasublenisacute"},
{"uni1F9D", "Etaiotasubasperacute"},
{"uni1F9E", "Etaiotasublenistilde"},
{"uni1F9F", "Etaiotasubaspertilde"},
{"uni1FA0", "omegaiotasublenis"},
{"uni1FA1", "omegaiotasubasper"},
{"uni1FA2", "omegaiotasublenisgrave"},
{"uni1FA3", "omegaiotasubaspergrave"},
{"uni1FA4", "omegaiotasublenisacute"},
{"uni1FA5", "omegaiotasubasperacute"},
{"uni1FA6", "omegaiotasublenistilde"},
{"uni1FA7", "omegaiotasubaspertilde"},
{"uni1FA8", "Omegaiotasublenis"},
{"uni1FA9", "Omegaiotasubasper"},
{"uni1FAA", "Omegaiotasublenisgrave"},
{"uni1FAB", "Omegaiotasubaspergrave"},
{"uni1FAC", "Omegaiotasublenisacute"},
{"uni1FAD", "Omegaiotasubasperacute"},
{"uni1FAE", "Omegaiotasublenistilde"},
{"uni1FAF", "Omegaiotasubaspertilde"},
{"uni1FB0", "alphabreve"},
{"uni1FB1", "alphamacron"},
{"uni1FB2", "alphaiotasubgrave"},
{"uni1FB3", "alphaiotasub"},
{"uni1FB4", "alphaiotasubacute"},
{"uni1FB6", "alphatilde"},
{"uni1FB7", "alphaiotasubtilde"},
{"uni1FB8", "Alphabreve"},
{"uni1FB9", "Alphamacron"},
{"uni1FBA", "Alphagrave"},
{"uni1FBB", "Alphaacute"},
{"uni1FBC", "Alphaiotasub"},
{"uni1FBD", "lenis"},
{"uni1FBE", "iotasubscript"},
{"uni1FC1", "dieresistilde"},
{"uni1FC2", "etaiotasubgrave"},
{"uni1FC3", "etaiotasub"},
{"uni1FC4", "etaiotasubacute"},
{"uni1FC6", "etatilde"},
{"uni1FC7", "etaiotasubtilde"},
{"uni1FC8", "Epsilongrave"},
{"uni1FC9", "Epsilonacute"},
{"uni1FCA", "Etagrave"},
{"uni1FCB", "Etaacute"},
{"uni1FCC", "Etaiotasub"},
{"uni1FCD", "lenisgrave"},
{"uni1FCE", "lenisacute"},
{"uni1FCF", "lenistilde"},
{"uni1FD0", "iotabreve"},
{"uni1FD1", "iotamacron"},
{"uni1FD2", "iotadieresisgrave"},
{"uni1FD3", "iotadieresisacute"},
{"uni1FD6", "iotatilde"},
{"uni1FD7", "iotadieresistilde"},
{"uni1FD8", "Iotabreve"},
{"uni1FD9", "Iotamacron"},
{"uni1FDA", "Iotagrave"},
{"uni1FDB", "Iotaacute"},
{"uni1FDD", "aspergrave"},
{"uni1FDE", "asperacute"},
{"uni1FDF", "aspertilde"},
{"uni1FE0", "upsilonbreve"},
{"uni1FE1", "upsilonmacron"},
{"uni1FE2", "upsilondieresisgrave"},
{"uni1FE3", "upsilondieresisacute"},
{"uni1FE4", "rholenis"},
{"uni1FE5", "rhoasper"},
{"uni1FE6", "upsilontilde"},
{"uni1FE7", "upsilondieresistilde"},
{"uni1FE8", "Upsilonbreve"},
{"uni1FE9", "Upsilonmacron"},
{"uni1FEA", "Upsilongrave"},
{"uni1FEB", "Upsilonacute"},
{"uni1FEC", "Rholenis"},
{"uni1FED", "dialytikagrave"},
{"uni1FED.cap", "dialytikagrave.cap"},
{"uni1FEE", "dialytikaacute"},
{"uni1FEE.cap", "dialytikaacute.cap"},
{"uni1FF2", "omegaiotasubgrave"},
{"uni1FF3", "omegaiotasub"},
{"uni1FF4", "omegaiotasubacute"},
{"uni1FF6", "omegatilde"},
{"uni1FF7", "omegaiotasubtilde"},
{"uni1FF8", "Omicrongrave"},
{"uni1FF9", "Omicronacute"},
{"uni1FFA", "Omegagrave"},
{"uni1FFB", "Omegaacute"},
{"uni1FFC", "Omegaiotasub"},
{"uni1FFE", "asper"},
{"uni2126", "Omega.math"},
{"uni2206", "Delta.math"},
{"uni2213", "minusplus"},
{"uniEFD5", "digamma"},
{"uniEFEE", "Breve"},
{"uniEFF1", "Ogonek"},
{"uniEFF2", "Cedilla"},
{"uniEFF3", "Ring"},
{"uniEFF5", "Tilde"},
{"uniEFF7", "Circumflex"},
{"uniF628", "perthousand.oldstyle"},
{"uniF629", "cent.denominator"},
{"uniF62A", "dollar.denominator"},
{"uniF62B", "hyphen.denominator"},
{"uniF62C", "parenleft.denominator"},
{"uniF62D", "parenright.denominator"},
{"uniF62E", "cent.numerator"},
{"uniF62F", "dollar.numerator"},
{"uniF630", "hyphen.numerator"},
{"uniF631", "parenleft.numerator"},
{"uniF632", "parenright.numerator"},
{"uniF633", "at.cap"},
{"uniF634", "commaaccent.cap"},
{"uniF635", "commaaccent.small"},
{"uniF636", "pi.math"},
{"uniF637", "h.superior"},
{"uniF638", "zero.slash"},
{"uniF639", "zero.fitted"},
{"uniF63A", "two.fitted"},
{"uniF63B", "three.fitted"},
{"uniF63C", "four.fitted"},
{"uniF63D", "five.fitted"},
{"uniF63E", "six.fitted"},
{"uniF63F", "seven.fitted"},
{"uniF640", "eight.fitted"},
{"uniF641", "nine.fitted"},
{"uniF642", "percent.oldstyle"},
{"uniF643", "zero.taboldstyle"},
{"uniF644", "one.taboldstyle"},
{"uniF645", "two.taboldstyle"},
{"uniF646", "three.taboldstyle"},
{"uniF647", "four.taboldstyle"},
{"uniF648", "five.taboldstyle"},
{"uniF649", "six.taboldstyle"},
{"uniF64A", "seven.taboldstyle"},
{"uniF64B", "eight.taboldstyle"},
{"uniF64C", "nine.taboldstyle"},
{"uniF64D", "colonmonetary.taboldstyle"},
{"uniF64E", "Euro.taboldstyle"},
{"uniF64F", "florin.taboldstyle"},
{"uniF650", "numbersign.taboldstyle"},
{"uniF651", "sterling.taboldstyle"},
{"uniF652", "yen.taboldstyle"},
{"uniF653", "dollar.taboldstyle"},
{"uniF654", "cent.taboldstyle"},
{"uniF655", "zero.denominator"},
{"uniF656", "one.denominator"},
{"uniF657", "two.denominator"},
{"uniF658", "three.denominator"},
{"uniF659", "four.denominator"},
{"uniF65A", "five.denominator"},
{"uniF65B", "six.denominator"},
{"uniF65C", "seven.denominator"},
{"uniF65D", "eight.denominator"},
{"uniF65E", "nine.denominator"},
{"uniF65F", "comma.denominator"},
{"uniF660", "period.denominator"},
{"uniF661", "zero.numerator"},
{"uniF662", "one.numerator"},
{"uniF663", "two.numerator"},
{"uniF664", "three.numerator"},
{"uniF665", "four.numerator"},
{"uniF666", "five.numerator"},
{"uniF667", "six.numerator"},
{"uniF668", "seven.numerator"},
{"uniF669", "eight.numerator"},
{"uniF66A", "nine.numerator"},
{"uniF66B", "comma.numerator"},
{"uniF66C", "period.numerator"},
{"uniF66D", "Abreve.small"},
{"uniF66E", "Amacron.small"},
{"uniF66F", "Aogonek.small"},
{"uniF670", "AEacute.small"},
{"uniF671", "Cacute.small"},
{"uniF672", "Ccaron.small"},
{"uniF673", "Ccircumflex.small"},
{"uniF674", "Cdotaccent.small"},
{"uniF675", "Dcaron.small"},
{"uniF676", "Dcroat.small"},
{"uniF677", "Ebreve.small"},
{"uniF678", "Ecaron.small"},
{"uniF679", "Edotaccent.small"},
{"uniF67A", "Emacron.small"},
{"uniF67B", "Eng.small"},
{"uniF67C", "Eogonek.small"},
{"uniF67D", "Gbreve.small"},
{"uniF67E", "Gcircumflex.small"},
{"uniF67F", "Gcommaaccent.small"},
{"uniF680", "Gdotaccent.small"},
{"uniF681", "Hbar.small"},
{"uniF682", "Hcircumflex.small"},
{"uniF683", "Ibreve.small"},
{"uniF684", "IJ.small"},
{"uniF685", "Imacron.small"},
{"uniF686", "Iogonek.small"},
{"uniF687", "Itilde.small"},
{"uniF688", "Jcircumflex.small"},
{"uniF689", "Kcommaaccent.small"},
{"uniF68A", "Lacute.small"},
{"uniF68B", "Lcaron.small"},
{"uniF68C", "Lcommaaccent.small"},
{"uniF68D", "Ldot.small"},
{"uniF68E", "Nacute.small"},
{"uniF68F", "Ncaron.small"},
{"uniF690", "Ncommaaccent.small"},
{"uniF691", "Obreve.small"},
{"uniF692", "Ohungarumlaut.small"},
{"uniF693", "Omacron.small"},
{"uniF694", "Oslashacute.small"},
{"uniF695", "Racute.small"},
{"uniF696", "Rcaron.small"},
{"uniF697", "Rcommaaccent.small"},
{"uniF698", "Sacute.small"},
{"uniF699", "Scedilla.small"},
{"uniF69A", "Scircumflex.small"},
{"uniF69B", "Scommaaccent.small"},
{"uniF69C", "Tbar.small"},
{"uniF69D", "Tcaron.small"},
{"uniF69E", "Tcommaaccent.small"},
{"uniF69F", "Ubreve.small"},
{"uniF6A0", "Uhungarumlaut.small"},
{"uniF6A1", "Umacron.small"},
{"uniF6A2", "Uogonek.small"},
{"uniF6A3", "Uring.small"},
{"uniF6A4", "Utilde.small"},
{"uniF6A5", "Wacute.small"},
{"uniF6A6", "Wcircumflex.small"},
{"uniF6A7", "Wdieresis.small"},
{"uniF6A8", "Wgrave.small"},
{"uniF6A9", "Ycircumflex.small"},
{"uniF6AA", "Ygrave.small"},
{"uniF6AB", "Zacute.small"},
{"uniF6AC", "Zdotaccent.small"},
{"uniF6AD", "Idotaccent.small"},
{"uniF6AE", "parenleft.cap"},
{"uniF6AF", "parenright.cap"},
{"uniF6B0", "bracketleft.cap"},
{"uniF6B1", "bracketright.cap"},
{"uniF6B2", "braceleft.cap"},
{"uniF6B3", "braceright.cap"},
{"uniF6B4", "exclamdown.cap"},
{"uniF6B5", "questiondown.cap"},
{"uniF6B6", "guillemotleft.cap"},
{"uniF6B7", "guillemotright.cap"},
{"uniF6B8", "guilsinglleft.cap"},
{"uniF6B9", "guilsinglright.cap"},
{"uniF6BA", "hyphen.cap"},
{"uniF6BB", "endash.cap"},
{"uniF6BC", "emdash.cap"},
{"uniF6BD", "periodcentered.cap"},
{"uniF6DC", "one.fitted"},
{"uniFB05", "longs_t"},
{"uniFB06", "s_t"},
