/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
   This software is licensed as OpenSource, under the Apache License, Version 2.0.
   This license is available at: http://opensource.org/licenses/Apache-2.0. */

/* svwerr library error name to error string mapping. See ctlshare.h for a
   description of the CTL_DCL_ERR macro. */

CTL_DCL_ERR(svwSuccess,       "no error")
CTL_DCL_ERR(svwErrNoMemory,   "out of memory")
CTL_DCL_ERR(svwErrDstStream,  "destination stream error")
CTL_DCL_ERR(svwErrTmpStream,  "temp stream error")
CTL_DCL_ERR(svwErrBadCall,    "inappropriate function call or arguments")
CTL_DCL_ERR(svwErrNotUnicode, "glyph Unicode value not provided")
CTL_DCL_ERR(svwErrFontName,   "FontName missing")
