"[-dcf options: defaults -T all, -5]\n"
"-T arg  select tables with arg, composed of the following letters:\n"
"    h   Header\n"
"    n   Name INDEX\n"
"    t   Top DICT INDEX\n"
"    s   String INDEX\n"
"    g   Global Subr INDEX\n"
"    e   Encoding\n"
"    C   Charset\n"
"    f   FDSelect\n"
"    F   FDArray\n"
"    c   CharStrings INDEX\n"
"    p   Private DICT\n"
"    l   Local Subr INDEX\n"
"   all  all tables selected (can't be combined with other letters)\n"
"-0      dump level 0\n"
"-1      dump level 1\n"
"-5      dump level 5\n",
"-b      don't break flowed objects\n"
"\n"
"Dump Compact Font (dcf) mode writes a literal dump of a Compact Font Format \n"
"(CFF) font. The form of the dump is controlled by the options above.\n"
"\n"
"The -T (table) option takes a single argument that is a string of 1 or more\n"
"table selection letters from the above list or the string \"all\" which will\n"
"select all tables. Only selected tables will be scheduled for dumping at the\n"
"level specified by the dump level options: -0, -1, and -5.\n"
"\n",
"The -0 (level 0) option simply lists the region a table occupies as an\n"
"inclusive start and end pair of absolute hexadecimal byte offsets.\n"
"\n"
"The -1 (level 1) option dumps all bytes of the table including structural data\n"
"like INDEX headers and offsets.\n"
"\n"
"The -5 (level 5) option leaves out the structural data that the -1 option\n"
"includes and some of the table data may be modified to make it easier to use.\n"
"\n"
"The -b (no break) option flows array-like object data onto a single line.\n"
"\n",
"There is no facility to map the Encoding or charset tables to glyph names since\n"
"-dump mode option -4 already provides it. When the String INDEX is dumped at\n"
"level 5 the indexes are converted to SIDs so that they can be used directly\n"
"with SID references in the Top DICT, Font DICT, or charset tables. When the\n"
"Global Subr INDEX or the Local Subr INDEX is dumped at level 5 the indexes are\n"
"biased so that they can be used directly with the callgsubr and callsubr\n"
"arguments.\n"
"\n",
"The charstring data in the Global Subr INDEX, Local Subr INDEX, and the\n"
"CharString INDEX is shown without flattening, i.e. the subroutine calls are not\n"
"interpreted. However, this may be overridden by using the -g option to select\n"
"glyphs. (Hint: you can select all the glyphs in the CharStrings INDEX by using\n"
"a range argument with the -g option, e.g. 0-N.) The CharStrings INDEX is\n"
"automatically selected by the -g option so selecting the table with the -T c\n"
"option isn't needed.\n"
"\n"
"For example, the command:\n"
"\n",
"    tx -dcf -g a BernhardModern-Roman.otf\n"
"\n"
"will write, on standard output, a flattened dump of the charstring data for\n"
"the glyph named \"a\" from the font file \"BernhardModern-Roman.otf\".\n"
