/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
   This software is licensed as OpenSource, under the Apache License, Version 2.0.
   This license is available at: http://opensource.org/licenses/Apache-2.0. */

/* cffread library error name to error string mapping. See ctlshare.h for
   description of CTL_DCL_ERR macro. */

CTL_DCL_ERR(pdwSuccess,      "no error")
CTL_DCL_ERR(pdwErrNoMemory,  "out of memory")
CTL_DCL_ERR(pdwErrSrcStream, "source stream error")
CTL_DCL_ERR(pdwErrDstStream, "destination stream error")
