/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
   This software is licensed as OpenSource, under the Apache License, Version 2.0.
   This license is available at: http://opensource.org/licenses/Apache-2.0. */

/* t1read library error name to error string mapping. See ctlshare.h for a
   description of the CTL_DCL_ERR macro. */

CTL_DCL_ERR(t1rSuccess,        "no error")
CTL_DCL_ERR(t1rErrCstrQuit,    "client quit glyph parse")
CTL_DCL_ERR(t1rErrCstrFail,    "client failed glyph parse")
CTL_DCL_ERR(t1rErrCstrParse,   "charstring parse error")
CTL_DCL_ERR(t1rErrNoMemory,    "out of memory")
CTL_DCL_ERR(t1rErrSrcStream,   "can't read source stream")
CTL_DCL_ERR(t1rErrTmpStream,   "tmp stream error")
CTL_DCL_ERR(t1rErrPostScript,  "PostScript parse error")
CTL_DCL_ERR(t1rErrEncoding,    "invalid encoding")
CTL_DCL_ERR(t1rErrCstrDecrypt, "can't decrypt charstring")
CTL_DCL_ERR(t1rErrUntermCstr,  "unterminated charstring")
CTL_DCL_ERR(t1rErrSubrsCount,  "invalid /Subrs count")
CTL_DCL_ERR(t1rErrSubrEntry,   "invalid subr entry")
CTL_DCL_ERR(t1rErrCharName,    "invalid char name")
CTL_DCL_ERR(t1rErrCharEntry,   "invalid char entry")
CTL_DCL_ERR(t1rErrKeyValue,    "invalid key value")
CTL_DCL_ERR(t1rErrFontType,    "unsupported /FontType")
CTL_DCL_ERR(t1rErrFDArray,     "too many FDArray elements or key redefined")
CTL_DCL_ERR(t1rErrCIDFontType, "unsupported /CIDFontType")
CTL_DCL_ERR(t1rErrChameleon,   "unsupported font type: Chameleon")
CTL_DCL_ERR(t1rErrHybrid,      "unsupported font type: hybrid")
CTL_DCL_ERR(t1rErrCIDMap,      "CIDMapOffset missing")
CTL_DCL_ERR(t1rErrSubrMap,     "SubrMapOffset missing")
CTL_DCL_ERR(t1rErrNoCID0,      "CID 0 (notdef) missing")
CTL_DCL_ERR(t1rErrStartData,   "unsupported StartData type")
CTL_DCL_ERR(t1rErrNoGlyph,     "requested glyph not in font")
CTL_DCL_ERR(t1rErrFontMatrix,  "invalid FontMatrix")
CTL_DCL_ERR(t1rErrBDP,         "BlendDesignPositions missing")
CTL_DCL_ERR(t1rErrBDM,         "BlendDesignMap missing")
CTL_DCL_ERR(t1rErrWV,          "WeightVector missing")
CTL_DCL_ERR(t1rErrGeometry,    "inconsistent MM geometry")
CTL_DCL_ERR(t1rErrCIDRange,    "/GlyphDirectory CID out of range")
CTL_DCL_ERR(t1rErrFDRange,     "FD index out of range")
CTL_DCL_ERR(t1rErrSubrLength,  "bad subr length")
CTL_DCL_ERR(t1rErrCharLength,  "bad charstring length")
CTL_DCL_ERR(t1rErrFontName,    "FontName missing")
CTL_DCL_ERR(t1rErrMMParse,     "Multiple Master parse error")
