/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
   This software is licensed as OpenSource, under the Apache License, Version 2.0.
   This license is available at: http://opensource.org/licenses/Apache-2.0. */

/* cfwerr library error name to error string mapping. See ctlshare.h for a
   description of the CTL_DCL_ERR macro. */

CTL_DCL_ERR(sfwSuccess, "no error")
CTL_DCL_ERR(sfwErrNoMemory, "out of memory")
CTL_DCL_ERR(sfwErrDstStream, "destination stream error")
CTL_DCL_ERR(sfwErrAbort, "client aborted function")
CTL_DCL_ERR(sfwErrDupTables, "duplicate tables registered")
CTL_DCL_ERR(sfwErrBadCall, "inappropriate function call or arguments")
