/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
This software is licensed as OpenSource, under the Apache License, Version 2.0. This license is available at: http://opensource.org/licenses/Apache-2.0. */

DCL_OPT("+E", opt__E)
DCL_OPT("+F", opt__F)
DCL_OPT("+O", opt__O)
DCL_OPT("+S", opt__S)
DCL_OPT("+T", opt__T)
DCL_OPT("+V", opt__V)
DCL_OPT("+Z", opt__Z)
DCL_OPT("+b", opt__b)
DCL_OPT("+d", opt__d)
DCL_OPT("+q", opt__q)
DCL_OPT("+w", opt__w)
DCL_OPT("-0", opt_0)
DCL_OPT("-1", opt_1)
DCL_OPT("-2", opt_2)
DCL_OPT("-3", opt_3)
DCL_OPT("-4", opt_4)
DCL_OPT("-5", opt_5)
DCL_OPT("-6", opt_6)
DCL_OPT("-A", opt_A)
DCL_OPT("-E", opt_E)
DCL_OPT("-F", opt_F)
DCL_OPT("-LWFN", opt_LWFN)
DCL_OPT("-N", opt_N)
DCL_OPT("-O", opt_O)
DCL_OPT("-P", opt_P)
DCL_OPT("-R", opt_R)
DCL_OPT("-S", opt_S)
DCL_OPT("-T", opt_T)
DCL_OPT("-U", opt_U)
DCL_OPT("-UNC", opt_UNC)
DCL_OPT("-V", opt_V)
DCL_OPT("-X", opt_X)
DCL_OPT("-Z", opt_Z)
DCL_OPT("-a", opt_a)
DCL_OPT("-abs", opt_abs)
DCL_OPT("-afm", opt_afm)
DCL_OPT("-altLayer", opt_altLayer)
DCL_OPT("-b", opt_b)
DCL_OPT("-bc", opt_bc)
DCL_OPT("-c", opt_c)
DCL_OPT("-cef", opt_cef)
DCL_OPT("-cefsvg", opt_cefsvg)
DCL_OPT("-cff", opt_cff)
DCL_OPT("-cff2", opt_cff2)
DCL_OPT("-cmp", opt_cmp)
DCL_OPT("-cr", opt_cr)
DCL_OPT("-crlf", opt_crlf)
DCL_OPT("-d", opt_d)
DCL_OPT("-dcf", opt_dcf)
DCL_OPT("-dd", opt_dd)
DCL_OPT("-decid", opt_decid)
DCL_OPT("-dump", opt_dump)
DCL_OPT("-e", opt_e)
DCL_OPT("-f", opt_f)
DCL_OPT("-fd", opt_fd)
DCL_OPT("-fdx", opt_fdx)
DCL_OPT("-g", opt_g)
DCL_OPT("-gn0", opt_gn0)
DCL_OPT("-gn1", opt_gn1)
DCL_OPT("-gn2", opt_gn2)
DCL_OPT("-gx", opt_gx)
DCL_OPT("-h", opt_h)
DCL_OPT("-i", opt_i)
DCL_OPT("-l", opt_l)
DCL_OPT("-lf", opt_lf)
DCL_OPT("-m", opt_m)
DCL_OPT("-maxs", opt_maxs)
DCL_OPT("-mtx", opt_mtx)
DCL_OPT("-n", opt_n)
DCL_OPT("-no_futile", opt_no_futile)
DCL_OPT("-no_opt", opt_no_opt)
DCL_OPT("-o", opt_o)
DCL_OPT("-p", opt_p)
DCL_OPT("-path", opt_path)
DCL_OPT("-pdf", opt_pdf)
DCL_OPT("-pfb", opt_pfb)
DCL_OPT("-pg", opt_pg)
DCL_OPT("-ps", opt_ps)
DCL_OPT("-q", opt_q)
DCL_OPT("-r", opt_r)
DCL_OPT("-s", opt_s)
DCL_OPT("-sa", opt_sa)
DCL_OPT("-sd", opt_sd)
DCL_OPT("-sha1", opt_sha1)
DCL_OPT("-sr", opt_sr)
DCL_OPT("-std", opt_std)
DCL_OPT("-svg", opt_svg)
DCL_OPT("-t", opt_t)
DCL_OPT("-t1", opt_t1)
DCL_OPT("-u", opt_u)
DCL_OPT("-ufo", opt_ufo)
DCL_OPT("-usefd", opt_usefd)
DCL_OPT("-v", opt_v)
DCL_OPT("-w", opt_w)
DCL_OPT("-x", opt_x)
DCL_OPT("-y", opt_y)
DCL_OPT("-z", opt_z)
DCL_OPT("/?", opt_h3)
DCL_OPT("/h", opt_h2)
DCL_OPT("/help", opt_h1)
