/* Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
   This software is licensed as OpenSource, under the Apache License, Version 2.0.
   This license is available at: http://opensource.org/licenses/Apache-2.0. */

/* pstoken library error name to error string mapping. See ctlshare.h for
   description of CTL_DCL_ERR macro. */

CTL_DCL_ERR(pstSuccess,       "no error")
CTL_DCL_ERR(pstErrNoMemory,   "no memory")
CTL_DCL_ERR(pstErrSrcStream,  "can't read source stream")
CTL_DCL_ERR(pstErrBadASCII85, "bad ASCII85 string")
CTL_DCL_ERR(pstErrBadHexStr,  "bad hex string")
CTL_DCL_ERR(pstErrBadCall,    "inappropriate function call or arguments")
CTL_DCL_ERR(pstErrMaxRecursion, "maximum recursion depth exceeded")
